from .FileTools.file_tools import (
    gen_folder_list,
    get_files,
    auto_split_md,
    auto_split_mds,
    unzips,
)
from .FileTools.extract_img import md_replace_imgs, mds_replace_imgs


__all__ = [
    "gen_folder_list",
    "get_files",
    "unzips",
    "md_replace_imgs",
    "mds_replace_imgs",
    "auto_split_md",
    "auto_split_mds",
]



LANGUAGES = ["简体中文", "Enlish"]
WORDS_CN = [
    "🔎 请选择文件预处理工具,其中Tool类代表其会直接处理PDF文件，OCR类代表其会仅会使用OCR处理图片文件：",
    "🔎 请选择备用文件预处理工具,其中Tool类代表其会直接处理PDF文件，OCR类代表其会仅会使用OCR处理图片文件：",
    "📇 请输入 Doc2X 的身份令牌，个人用户请访问 https://doc2x.noedgeai.com/ 获取：",
    "⚠️ 验证 Doc2X 的身份令牌失败，请检查网络连接或者身份令牌是否正确",
    "📌 请选择 Doc2X 的速率限制，含意为同时请求数量，建议输入 A 以自动选择速率限制：",
]
WORDS_EN = [
    "🔎 Please select the file pre-processing tool, where the Tool class represents that it will directly process PDF files, and the OCR class represents that it will only use OCR to process image files:",
    "🔎 Please select the fallback file pre-processing tool, where the Tool class represents that it will directly process PDF files, and the OCR class represents that it will only use OCR to process image files:",
    "📇 Please enter the API key of the Doc2X, for personal use, visit https://doc2x.com/ to get the key:",
    "⚠️ Failed to verify the API key of Doc2X, please check the network connection or the API key",
    "📌 Please select the rate limit of Doc2X, means number of simultaneous requests, it is recommended to enter A to automatically select the rate limit:",
]
WORDS = [WORDS_CN, WORDS_EN]


WORDS_LOCAL_CN = [
    "📂 请输入项目文件夹的名称：",
    "⚠️ 文件夹已存在，请重新输入：",
    "🔗 正在尝试在以下路径初始化项目文件夹配置：",
    "💾 找到全局配置，是否使用全局配置？",
    "🌐 请选择预处理完成后的文件输出方式：",
]
WORDS_LOCAL_EN = [
    "📂 Please enter the name of the project folder:",
    "⚠️ The folder already exists, please re-enter:",
    "🔗 Trying to initialize the project folder configuration at the following path:",
    "💾 Found global configuration, do you want to use it?",
    "🌐 Please select the file output method after preprocessing:",
]

WORDS_LOCAL = [WORDS_LOCAL_CN, WORDS_LOCAL_EN]



import httpx
import json
import os
import re
from typing import Tuple
from .Exception import RateLimit, FileError, RequestError, async_retry, code_check
import logging
from .Types import OutputFormat

Base_URL = "https://v2.doc2x.noedgeai.com/api"

logger = logging.getLogger("pdfdeal.convertV2")


@async_retry()
async def upload_pdf(apikey: str, pdffile: str, ocr: bool = True) -> str:
    """Upload pdf file to server and return the uid of the file

    Args:
        apikey (str): The key
        pdffile (str): The pdf file path
        ocr (bool, optional): Do OCR or not. Defaults to True.

    Raises:
        FileError: Input file size is too large
        FileError: Open file error
        RateLimit: Rate limit exceeded
        Exception: Upload file error

    Returns:
        str: The uid of the file
    """
    url = f"{Base_URL}/v2/parse/pdf"
    if os.path.getsize(pdffile) >= 300 * 1024 * 1024:
        logger.warning("Now not support PDF file > 300MB!")
        raise RequestError("parse_file_too_large")
        logger.warning(
            "File size is too large, will auto switch to S3 file upload way, this may take a while"
        )
        return await upload_pdf_big(apikey, pdffile, ocr)

    try:
        with open(pdffile, "rb") as f:
            file = f.read()
    except Exception as e:
        raise FileError(f"Open file error! {e}")

    async with httpx.AsyncClient(timeout=httpx.Timeout(120)) as client:
        post_res = await client.post(
            url,
            params={"ocr": str(ocr).lower()},
            headers={
                "Authorization": f"Bearer {apikey}",
                "Content-Type": "application/pdf",
            },
            content=file,
        )

    if post_res.status_code == 200:
        response_data = json.loads(post_res.content.decode("utf-8"))
        uid = response_data.get("data", {}).get("uid")
        trace_id = post_res.headers.get("trace-id")
        await code_check(
            code=response_data.get("code", response_data), uid=uid, trace_id=trace_id
        )
        return uid

    if post_res.status_code == 429:
        raise RateLimit()
    if post_res.status_code == 400:
        raise RequestError(post_res.text)

    raise Exception(f"Upload file error! {post_res.status_code}:{post_res.text}")


@async_retry()
async def upload_pdf_big(apikey: str, pdffile: str, ocr: bool = True) -> str:
    """Upload big pdf file(>100m) to server and return the uid of the file

    Args:
        apikey (str): The key
        pdffile (str): The pdf file path
        ocr (bool, optional): Do OCR or not. Defaults to True.

    Raises:
        FileError: Input file size is too large
        FileError: Open file error
        RateLimit: Rate limit exceeded
        Exception: Upload file error

    Returns:
        str: The uid of the file
    """
    if os.path.getsize(pdffile) < 100 * 1024 * 1024:
        raise FileError("PDF file size should be bigger than 100MB!")

    try:
        file = {"file": open(pdffile, "rb")}
    except Exception as e:
        raise FileError(f"Open file error! {e}")

    url = f"{Base_URL}/v2/parse/preupload"
    filename = os.path.basename(pdffile)[:20]
    timeout = httpx.Timeout(120)

    async with httpx.AsyncClient(timeout=timeout) as client:
        post_res = await client.post(
            url,
            headers={"Authorization": f"Bearer {apikey}"},
            json={"file_name": filename, "ocr": ocr},
        )

    if post_res.status_code == 200:
        response_data = json.loads(post_res.content.decode("utf-8"))
        uid = response_data["data"]["form"]["x-amz-meta-uid"]
        trace_id = post_res.headers.get("trace-id")
        await code_check(
            response_data.get("code", response_data), uid=uid, trace_id=trace_id
        )

        upload_data = response_data["data"]
        upload_url = upload_data["url"]

        async with httpx.AsyncClient(timeout=timeout) as client:
            s3_res = await client.post(
                url=upload_url,
                data=upload_data["form"],
                files=file,
            )
        if s3_res.status_code == 204:
            return uid
        else:
            raise RequestError(s3_res.text)
    if post_res.status_code == 400:
        raise RequestError(post_res.text)
    raise Exception(f"Upload file error! {post_res.status_code}:{post_res.text}")


async def decode_data(data: dict, convert: bool) -> Tuple[list, list]:
    """Decode the data

    Args:
        data (dict): The response data
        convert (bool): Convert "[" and "[[" to "$" and "$$"

    Returns:
        Tuple[list, list]: The texts and locations
    """
    texts = []
    locations = []
    if "result" not in data or "pages" not in data["result"]:
        logger.warning("Although parsed successfully, the content is empty!")
        return [], []
    texts, locations = [], []
    for page in data["result"]["pages"]:
        text = page.get("md", "")
        if convert:
            text = re.sub(r"\\[()]", "$", text)
            text = re.sub(r"\\[\[\]]", "$$", text)
        texts.append(text)
        locations.append(
            {
                "url": page.get("url", ""),
                "page_idx": page.get("page_idx", 0),
                "page_width": page.get("page_width", 0),
                "page_height": page.get("page_height", 0),
            }
        )
    return texts, locations


@async_retry()
async def uid_status(
    apikey: str,
    uid: str,
    convert: bool = False,
) -> Tuple[int, str, list, list]:
    """Get the status of the file

    Args:
        apikey (str): The key
        uid (str): The uid of the file
        convert (bool, optional): Convert "[" and "[[" to "$" and "$$" or not. Defaults to False.

    Raises:
        RequestError: Failed to deal with file
        Exception: Get status error

    Returns:
        Tuple[int, str, list, list]: The progress, status, texts and locations
    """
    url = f"{Base_URL}/v2/parse/status?uid={uid}"
    async with httpx.AsyncClient(timeout=httpx.Timeout(30)) as client:
        response_data = await client.get(
            url, headers={"Authorization": f"Bearer {apikey}"}
        )
    if response_data.status_code != 200:
        raise Exception(
            f"Get status error! {response_data.status_code}:{response_data.text}"
        )

    try:
        data = json.loads(response_data.content.decode("utf-8"))
    except Exception as e:
        raise Exception(
            f"Get status error with {e}! {response_data.status_code}:{response_data.text}"
        )

    trace_id = response_data.headers.get("trace-id")
    await code_check(data.get("code", response_data), uid, trace_id=trace_id)

    progress, status = data["data"].get("progress", 0), data["data"].get("status", "")
    if status == "processing":
        return progress, "Processing file", [], []
    elif status == "success":
        texts, locations = await decode_data(data["data"], convert)
        return 100, "Success", texts, locations
    elif status == "failed":
        raise RequestError(f"Failed to deal with file! {response_data.text}")
    else:
        logger.warning(f"Unknown status: {status}")
        return progress, status, [], []


@async_retry()
async def convert_parse(
    apikey: str, uid: str, to: str, filename: str = None
) -> Tuple[str, str]:
    """Convert parsed file to specified format

    Args:
        apikey (str): The API key
        uid (str): The uid of the parsed file
        to (str): Export format, supports: md|tex|docx|md_dollar
        filename (str, optional): Output filename for md/tex (without extension). Defaults to None.

    Raises:
        ValueError: If 'to' is not a valid format
        RequestError: If the conversion fails
        Exception: For any other errors during the process

    Returns:
        Tuple[str, str]: A tuple containing the status and URL of the converted file
    """
    url = f"{Base_URL}/v2/convert/parse"

    to = OutputFormat(to)
    if isinstance(to, OutputFormat):
        to = to.value

    payload = {"uid": uid, "to": to, "formula_mode": "normal"}
    if filename and to in ["md", "md_dollar", "tex"]:
        payload["filename"] = filename
    if to == "md_dollar":
        payload["formula_mode"] = "dollar"
        payload["to"] = "md"
    async with httpx.AsyncClient(timeout=httpx.Timeout(30)) as client:
        response_data = await client.post(
            url, json=payload, headers={"Authorization": f"Bearer {apikey}"}
        )

    if response_data.status_code != 200:
        raise Exception(
            f"Conversion request failed: {response_data.status_code}:{response_data.text}"
        )

    data = response_data.json()
    trace_id = response_data.headers.get("trace-id")
    await code_check(data.get("code", response_data), uid, trace_id=trace_id)
    status = data["data"]["status"]
    url = data["data"].get("url", "")

    if status == "processing":
        return "Processing", ""
    elif status == "success":
        return "Success", url
    else:
        raise RequestError(f"Conversion uid {uid} file failed: {data}")


@async_retry()
async def get_convert_result(apikey: str, uid: str) -> Tuple[str, str]:
    """Get the result of a conversion task

    Args:
        apikey (str): The API key
        uid (str): The uid of the conversion task

    Raises:
        RequestError: If the request fails
        Exception: For any other errors during the process

    Returns:
        Tuple[str, str]: A tuple containing the status and URL of the converted file
    """
    url = f"{Base_URL}/v2/convert/parse/result"

    params = {"uid": uid}

    async with httpx.AsyncClient(timeout=httpx.Timeout(30)) as client:
        response = await client.get(
            url, params=params, headers={"Authorization": f"Bearer {apikey}"}
        )

    if response.status_code != 200:
        raise Exception(
            f"Get conversion result failed: {response.status_code}:{response.text}"
        )

    data = response.json()
    trace_id = response.headers.get("trace-id")
    await code_check(data.get("code", response), uid, trace_id=trace_id)
    status = data["data"]["status"]
    url = data["data"].get("url", "")

    if status == "processing":
        return "Processing", ""
    elif status == "success":
        return "Success", url
    else:
        raise RequestError(f"Get conversion result for uid {uid} failed: {data}")


@async_retry()
async def download_file(
    url: str, file_type: str, target_folder: str, target_filename: str
) -> str:
    """
    Download a file from the given URL to the specified target folder with the given filename.

    Args:
        url (str): The URL to download the file from.
        file_type (str): The type of file being downloaded (e.g., 'zip', 'docx').
        target_folder (str): The folder where the file should be saved.
        target_filename (str): The desired filename for the downloaded file, can include subdirectories.

    Raises:
        Exception: If there's an error creating the target folder or downloading the file.

    Returns:
        str: The full path of the downloaded file.
    """
    target_path = os.path.join(target_folder, target_filename)
    target_dir = os.path.dirname(target_path)
    filename = os.path.basename(target_path)
    os.makedirs(target_dir, exist_ok=True)

    filename = os.path.splitext(filename)[0]
    if file_type != "docx":
        file_type = "zip"
    file_path = os.path.join(target_dir, f"{filename}.{file_type}")
    counter = 1
    while os.path.exists(file_path):
        file_path = os.path.join(target_dir, f"{filename}_{counter}.{file_type}")
        counter += 1

    async with httpx.AsyncClient(timeout=httpx.Timeout(60)) as client:
        response = await client.get(url)
        response.raise_for_status()
        with open(file_path, "wb") as f:
            f.write(response.content)

    return file_path



# from pdfdeal import Doc2X
# from pdfdeal.file_tools import get_files
# import os
# import logging

# httpx_logger = logging.getLogger("httpx")
# httpx_logger.setLevel(logging.WARNING)
# logging.basicConfig(level=logging.INFO)


# def test_single_pic2file():
#     client = Doc2X()
#     filepath, _, _ = client.pic2file(
#         image_file="tests/image/sample.png",
#         output_path="./Output/test/single/pic2file",
#         output_names=["pic_sample1.docx"],
#         output_format="docx",
#     )
#     if filepath[0] != "":
#         assert os.path.exists(filepath[0])
#         assert os.path.isfile(filepath[0])
#         assert filepath[0].endswith(".docx")
#         assert os.path.basename(filepath[0]) == "pic_sample1.docx"


# def test_multiple_pic2file():
#     client = Doc2X()
#     file_list, rename = get_files("tests/image", "img", "docx")
#     success, failed, flag = client.pic2file(
#         image_file=file_list,
#         output_path="./Output/test/multiple/pic2file",
#         output_names=rename,
#         output_format="docx",
#     )
#     assert flag
#     assert len(success) == len(failed) == 3
#     for s in success:
#         if s != "":
#             assert s.endswith("sample1.docx") or s.endswith("sample.docx")


# # def test_multiple_high_rpm():
# #     client = Doc2X()
# #     file_list = ["tests/image/sample.png" for _ in range(20)]
# #     success, failed, flag = client.pic2file(
# #         image_file=file_list,
# #         output_path="./Output/test",
# #     )
# #     assert len(success) == len(failed) == 20
# #     i = 0
# #     for s in success:
# #         if s != "":
# #             assert s.endswith(".zip")
# #         else:
# #             i += 1
# #     print(f"===Failed {i} times===")



import oss2
import logging


class OSS:
    def __init__(self, OSS_ACCESS_KEY_ID, OSS_ACCESS_KEY_SECRET, Endpoint, Bucket):
        """Initialize the OSS client.

        Args:
            OSS_ACCESS_KEY_ID (str): The access key ID for Aliyun OSS.
            OSS_ACCESS_KEY_SECRET (str): The access key secret for Aliyun OSS.
            Endpoint (str): The endpoint for Aliyun OSS.
            Bucket (str): The name of the bucket to upload to.
        """
        self.auth = oss2.Auth(OSS_ACCESS_KEY_ID, OSS_ACCESS_KEY_SECRET)
        self.bucket = oss2.Bucket(self.auth, Endpoint, Bucket)

    def upload_file(self, local_file_path, remote_file_path):
        """Upload a file to Aliyun OSS.

        Args:
            local_file_path (str): The path of the local file to upload.
            remote_file_path (str): The path of the remote file to upload to.

        Returns:
            tuple: A tuple containing the URL of the uploaded file and a boolean indicating whether the upload was successful.
        """
        try:
            self.bucket.put_object_from_file(remote_file_path, local_file_path)
            # * Default think the bucket is public read
            return (
                f"https://{self.bucket.bucket_name}.{self.bucket.endpoint.split('://')[1]}/{remote_file_path}",
                True,
            )
        except Exception as e:
            logging.error(f"Error to upload the file: {local_file_path}, {e}")
            return e, False


def Ali_OSS(OSS_ACCESS_KEY_ID, OSS_ACCESS_KEY_SECRET, Endpoint, Bucket) -> callable:
    """Initialize the OSS client and return a callable function to upload files.

    Args:
        OSS_ACCESS_KEY_ID (str): The access key ID for Aliyun OSS.
        OSS_ACCESS_KEY_SECRET (str): The access key secret for Aliyun OSS.
        Endpoint (str): The endpoint for Aliyun OSS.
        Bucket (str): The name of the bucket to upload to.

    Returns:
        callable: The upload_file method of the OSS client.
    """
    ali_oss = OSS(OSS_ACCESS_KEY_ID, OSS_ACCESS_KEY_SECRET, Endpoint, Bucket)
    return ali_oss.upload_file



import logging.config
from pdfdeal import Doc2X
import logging

httpx_logger = logging.getLogger("httpx")
httpx_logger.setLevel(logging.WARNING)
logging.basicConfig(level=logging.INFO)


def test_client():
    client = Doc2X()
    assert client is not None



def strore_pdf(pdf_path, Text):
    from reportlab.lib.pagesizes import letter
    from reportlab.pdfgen import canvas
    from reportlab.pdfbase.cidfonts import UnicodeCIDFont
    from reportlab.pdfbase import pdfmetrics

    c = canvas.Canvas(pdf_path, pagesize=letter)

    pdfmetrics.registerFont(UnicodeCIDFont("STSong-Light"))
    c.setFont("STSong-Light", 12)

    for text in Text:
        lines = text.split("\n")
        for i, line in enumerate(lines):
            c.setFont("STSong-Light", 12)
            c.drawString(100, 750 - i * 13, line)
        c.showPage()
    c.save()



import re
from typing import Tuple


def gen_imglist_from_md(mdfile: str) -> Tuple[list, list]:
    """
    Find the images used in the markdown file.
    Args:
        `mdfile`: `str`, the markdown file path.
    Returns:
        `imglist`: `list`, the image texts list used in the markdown file.
        `imgpath`: `list`, the image path list used in the markdown file.
    """
    imglist = []
    imgpath = []

    with open(mdfile, "r", encoding="utf-8") as file:
        content = file.read()

    # Get ![alt text](image_path) or <img src="image_path" alt="alt text">
    pattern = r'!\[([^\]]*)\]\(([^)]+)\)|<img\s+src="([^"]+)"\s+alt="([^"]*)">'
    matches = re.findall(pattern, content)

    for match in matches:
        if match[0] and match[1]:  # The type ![alt text](image_path)
            imglist.append(f"![{match[0]}]({match[1]})")
            imgpath.append(match[1])
        elif match[2] and match[3]:  # The type <img src="image_path" alt="alt text">
            imglist.append(f'<img src="{match[2]}" alt="{match[3]}">')
            imgpath.append(match[2])

    return imglist, imgpath


def split_of_md(mdfile: str, mode: str) -> list:
    """Find the header of the markdown file and add a split line before it.

    Args:
        mdfile (str): The markdown file path.
        mode (str): The way to split. Only support `title`(split by every title) now.

    Returns:
        list: A list of strings where each string is a segment of the markdown file between headers.
    """
    with open(mdfile, "r", encoding="utf-8") as file:
        content = file.read()

    pattern = r"(^#{1,6}\s.*$)"
    matches = re.finditer(pattern, content, re.MULTILINE)
    headers = [match.group() for match in matches]

    segments = []
    start = 0
    for header in headers:
        end = content.find(header, start)
        segments.append(content[start:end].strip())
        start = end

    segments.append(content[start:].strip())
    return segments



import asyncio
import os
from typing import Tuple, List
import logging
from .Doc2X.ConvertV2 import (
    upload_pdf,
    uid_status,
    convert_parse,
    get_convert_result,
    download_file,
)
from .Doc2X.Types import OutputFormat
from .Doc2X.Pages import get_pdf_page_count
from .Doc2X.Exception import RequestError, RateLimit, run_async
from .FileTools.file_tools import get_files

logger = logging.getLogger(name="pdfdeal.doc2x")


async def pdf2file(
    apikey: str,
    pdf_path: str,
    output_path: str,
    output_format: str,
    output_name: str,
    ocr: bool,
    maxretry: int,
    wait_time: int,
    max_time: int,
    convert: bool,
) -> str:
    """Convert pdf file to specified file using V2 API"""

    async def retry_upload():
        for _ in range(maxretry):
            try:
                return await upload_pdf(apikey, pdf_path, ocr)
            except RateLimit:
                logger.warning("Rate limit reached, retrying...")
                await asyncio.sleep(wait_time)
        raise RequestError("Max retry reached for upload_pdf")

    logger.info(f"Uploading {pdf_path}...")
    try:
        uid = await upload_pdf(apikey, pdf_path, ocr)
    except RateLimit:
        uid = await retry_upload()

    for _ in range(max_time):
        progress, status, texts, locations = await uid_status(apikey, uid, convert)
        if status == "Success":
            logger.info(f"Conversion successful for {pdf_path} with uid {uid}")
            if output_format == "texts":
                return texts
            elif output_format == "text":
                return "\n".join(texts)
            elif output_format == "detailed":
                return [
                    {"text": text, "location": loc}
                    for text, loc in zip(texts, locations)
                ]
            elif output_format in ["md", "md_dollar", "tex", "docx"]:
                logger.info(f"Parsing {uid} to {output_format}...")
                status, url = await convert_parse(apikey, uid, output_format)
                for _ in range(max_time):
                    if status == "Success":
                        logger.info(
                            f"Downloading {uid} {output_format} file to {output_path}..."
                        )
                        return await download_file(
                            url=url,
                            file_type=output_format,
                            target_folder=output_path,
                            target_filename=output_name or uid,
                        )
                    elif status == "Processing":
                        await asyncio.sleep(1)
                        status, url = await get_convert_result(apikey, uid)
                    else:
                        raise RequestError(f"Unexpected status: {status}")
                raise RequestError("Max time reached for get_convert_result")
            else:
                raise ValueError(f"Unsupported output format: {output_format}")
        elif status == "Processing file":
            logger.info(f"Processing {uid} : {progress}%")
            await asyncio.sleep(1)
        else:
            raise RequestError(f"Unexpected status: {status}")
    raise RequestError("Max time reached for uid_status")


class Doc2X:
    def __init__(
        self,
        apikey: str = None,
        thread: int = 5,
        max_pages: int = 1000,
        retry_time: int = 3,
        max_time: int = 90,
        debug: bool = False,
    ) -> None:
        """
        Initialize a Doc2X client.

        Args:
            apikey (str, optional): The API key for Doc2X. If not provided, it will try to get from environment variable 'DOC2X_APIKEY'.
            thread (int, optional): The maximum number of concurrent threads. Defaults to 5.
            max_pages (int, optional): The maximum number of pages to process. Defaults to 1000.
            retry_time (int, optional): The number of retry attempts. Defaults to 2.
            max_time (int, optional): The maximum time (in seconds) to wait for a response. Defaults to 90.
            debug (bool, optional): Whether to enable debug logging. Defaults to False.

        Raises:
            ValueError: If no API key is found.

        Attributes:
            apikey (str): The API key for Doc2X.
            retry_time (int): The number of retry attempts.
            max_time (int): The maximum time to wait for a response.
            thread (int): The maximum number of concurrent threads.
            max_pages (int): The maximum number of pages to process.
            debug (bool): Whether debug logging is enabled.

        Note:
            If debug is set to True, it will set the logging level of 'pdfdeal' logger to DEBUG.
        """
        self.apikey = apikey or os.environ.get("DOC2X_APIKEY", "")
        if not self.apikey:
            raise ValueError("No apikey found")
        self.retry_time = retry_time
        self.max_time = max_time
        self.thread = thread
        self.max_pages = max_pages

        handler = logging.StreamHandler()
        formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
        handler.setFormatter(formatter)
        logger.handlers.clear()
        logger.addHandler(handler)
        logger.propagate = False
        if debug:
            logging.getLogger("pdfdeal").setLevel(logging.DEBUG)
        self.debug = debug

    async def pdf2file_back(
        self,
        pdf_file,
        output_names: List[str] = None,
        output_path: str = "./Output",
        output_format: str = "md_dollar",
        ocr: bool = True,
        convert: bool = False,
    ) -> Tuple[List[str], List[dict], bool]:
        if isinstance(pdf_file, str):
            if os.path.isdir(pdf_file):
                pdf_file, output_names = get_files(
                    path=pdf_file, mode="pdf", out=output_format
                )
            else:
                pdf_file = [pdf_file]
                if output_names is None:
                    output_names = [os.path.basename(pdf_file[0])]

        output_names = output_names or [None] * len(pdf_file)
        if len(pdf_file) != len(output_names):
            raise ValueError("The length of files and output_names should be the same.")

        try:
            output_format = OutputFormat(output_format)
        except ValueError as e:
            raise ValueError(f"Invalid output format: {e}")

        output_format = output_format.value

        semaphore = asyncio.Semaphore(self.max_pages)
        thread_semaphore = asyncio.Semaphore(self.thread)

        async def process_file(pdf, name):
            async with thread_semaphore:
                try:
                    page_count = get_pdf_page_count(pdf)
                except Exception as e:
                    logger.warning(f"Failed to get page count for {pdf}: {str(e)}")
                    page_count = self.max_pages - 1  #! Assume the worst case
                if page_count > self.max_pages:
                    logger.warning(f"File {pdf} has too many pages, skipping.")
                    raise ValueError(f"File {pdf} has too many pages.")

                async def acquire_semaphore():
                    for _ in range(page_count):
                        await semaphore.acquire()

                try:
                    await acquire_semaphore()
                    try:
                        result = await asyncio.wait_for(
                            pdf2file(
                                apikey=self.apikey,
                                pdf_path=pdf,
                                output_path=os.path.join(output_path),
                                output_format=output_format,
                                output_name=name,
                                ocr=ocr,
                                wait_time=15,
                                max_time=self.max_time,
                                maxretry=self.retry_time,
                                convert=convert,
                            ),
                            timeout=self.max_time * 2,
                        )
                        return result, "", True
                    except asyncio.TimeoutError:
                        return "", "Operation timed out", False
                    except Exception as e:
                        return "", str(e), False
                finally:
                    #! Assuming that the semaphore release should be done regardless of the outcome
                    for _ in range(page_count):
                        semaphore.release()

        results = await asyncio.gather(
            *[process_file(pdf, name) for pdf, name in zip(pdf_file, output_names)]
        )

        success_files = [r[0] if r[2] else "" for r in results]
        failed_files = [
            {"error": r[1], "path": pdf} if not r[2] else {"error": "", "path": ""}
            for r, pdf in zip(results, pdf_file)
        ]
        has_error = any(not r[2] for r in results)

        if has_error:
            failed_count = sum(1 for fail in failed_files if fail["error"] != "")
            logger.error(
                f"{failed_count} file(s) failed to convert, please enable DEBUG mod to check or read the output variable."
            )
            if self.debug:
                for fail in failed_files:
                    if fail["error"] != "":
                        print("====================================")
                        print(f"Failed to convert {fail['path']}: {fail['error']}")
        logger.info(
            f"Successfully converted {sum(1 for file in success_files if file)} file(s)."
        )
        return success_files, failed_files, has_error

    def pdf2file(
        self,
        pdf_file,
        output_names: List[str] = None,
        output_path: str = "./Output",
        output_format: str = "md_dollar",
        ocr: bool = True,
        convert: bool = False,
        retry: bool = False,
    ) -> Tuple[List[str], List[dict], bool]:
        """Convert PDF files to the specified format.

        Args:
            pdf_file (str | List[str]): Path to a single PDF file or a list of PDF file paths.
            output_names (List[str], optional): List of output file names. Defaults to None.
            output_path (str, optional): Directory path for output files. Defaults to "./Output".
            output_format (str, optional): Desired output format. Defaults to `md_dollar`. Supported formats include:`md_dollar`|`md`|`tex`|`docx`, support output variable: `txt`|`txts`|`detailed`
            ocr (bool, optional): Whether to use OCR. Defaults to True.
            convert (bool, optional): Whether to convert Convert "[" and "[[" to "$" and "$$", only valid if `output_format` is a variable format(`txt`|`txts`|`detailed`). Defaults to False.
            retry (bool, optional): Whether to retry failed conversions. Defaults to False.

        Returns:
            Tuple[List[str], List[dict], bool]: A tuple containing:
                1. A list of successfully converted file paths or content.
                2. A list of dictionaries containing error information for failed conversions.
                3. A boolean indicating whether any errors occurred during the conversion process.

        Raises:
            Any exceptions raised by pdf2file_back or run_async.

        Note:
            This method provides a convenient synchronous interface for the asynchronous
            PDF conversion functionality. It handles all the necessary setup for running
            the asynchronous code in a synchronous context.
        """
        success_files, failed_files, has_error = run_async(
            self.pdf2file_back(
                pdf_file=pdf_file,
                output_names=output_names,
                output_path=output_path,
                output_format=output_format,
                ocr=ocr,
                convert=convert,
            )
        )

        if retry and has_error:
            logger.warning(
                "Detected failed conversions, retrying... If your file was parsed incorrectly, there will be no additional charges"
            )
            logger.warning(
                "Warning: If your file was exported in error, this may incur an additional charge!"
            )
            retry_pdf_files = [f["path"] for f in failed_files if f["path"]]
            retry_output_names = [os.path.basename(f) for f in retry_pdf_files]
            retry_success, retry_failed, retry_has_error = run_async(
                self.pdf2file_back(
                    pdf_file=retry_pdf_files,
                    output_names=retry_output_names,
                    output_path=output_path,
                    output_format=output_format,
                    ocr=ocr,
                    convert=convert,
                )
            )

            # Merge retry results with original results
            for i, retry_file in enumerate(retry_pdf_files):
                original_index = next(
                    index
                    for index, f in enumerate(failed_files)
                    if f["path"] == retry_file
                )
                if retry_success[i]:
                    success_files.insert(original_index, retry_success[i])
                else:
                    failed_files[original_index] = retry_failed[i]

            failed_files = [
                f
                for f in failed_files
                if f["path"] not in retry_pdf_files
                or not retry_success[retry_pdf_files.index(f["path"])]
            ]
            has_error = any(f["error"] for f in failed_files)

        return success_files, failed_files, has_error



from pypdf import PdfReader
from .Exception import RequestError


def get_pdf_page_count(pdf_path: str) -> int:
    """
    Get the number of pages in a PDF file.

    Args:
        pdf_path (str): The path to the PDF file.

    Returns:
        int: The number of pages in the PDF.
    """
    with open(pdf_path, "rb") as file:
        reader = PdfReader(file)
        pages = len(reader.pages)
    if pages > 1000:
        raise RequestError("parse_file_page_limit")
    return pages



from pdfdeal import Doc2X
import os
import pytest


def test_single_pdf2file():
    client = Doc2X(debug=True, thread=1)
    filepath, failed, flag = client.pdf2file(
        pdf_file="tests/pdf/sample.pdf",
        output_path="./Output/test/single/pdf2file",
        output_names=["Test.zip"],
        output_format="md_dollar",
    )
    if filepath[0] != "":
        assert os.path.exists(filepath[0])
        assert os.path.isfile(filepath[0])
        assert filepath[0].endswith(".zip")
        assert os.path.basename(filepath[0]) == "Test.zip"
    print(filepath)
    print(failed)
    print(flag)


def test_error_input_pdf2file():
    client = Doc2X(debug=True, thread=1)
    with pytest.raises(ValueError):
        client.pdf2file(
            pdf_file="tests/pdf/sample.pdf",
            output_path="./Output/test/single/pdf2file",
            output_names=["sample1.zip"],
            output_format="md_dallar",
        )


def test_multiple_pdf2file():
    client = Doc2X(debug=True, thread=1)
    success, failed, flag = client.pdf2file(
        pdf_file="tests/pdf",
        output_path="./Output/test/multiple/pdf2file",
        output_format="docx",
    )
    assert flag
    assert len(success) == len(failed) == 3
    for s in success:
        if s != "":
            assert s.endswith("sample.docx") or s.endswith("sampleB.docx")
    print(success)
    print(failed)
    print(flag)



from enum import Enum


class OutputFormat(str, Enum):
    DOCX = "docx"
    TXTS = "texts"
    TXT = "text"
    DETAILED = "detailed"
    LATEX = "tex"
    MD = "md"
    MD_DOLLAR = "md_dollar"

    @classmethod
    def _missing_(cls, value):
        for member in cls:
            if member.value.lower() == value.lower():
                return member
        raise ValueError(
            f"{value} is not a valid {cls.__name__}, must be one of {', '.join([m.value for m in cls])}"
        )


class OutputFormat_Legacy(str, Enum):
    DOCX = "docx"
    TXTS = "texts"
    LATEX = "latex"
    MD = "md"
    MD_DOLLAR = "md_dollar"

    @classmethod
    def _missing_(cls, value):
        for member in cls:
            if member.value.lower() == value.lower():
                return member
        raise ValueError(
            f"{value} is not a valid {cls.__name__}, must be one of {', '.join([m.value for m in cls])}"
        )


class RAG_OutputType(str, Enum):
    PDF = "pdf"
    MD = "md"
    TEXTS = "texts"

    @classmethod
    def _missing_(cls, value):
        for member in cls:
            if member.value.lower() == value.lower():
                return member
        raise ValueError(
            f"{value} is not a valid {cls.__name__}, must be one of {', '.join([m.value for m in cls])}"
        )


class Support_File_Type(str, Enum):
    PDF = "pdf"
    IMG = "img"

    @classmethod
    def _missing_(cls, value):
        for member in cls:
            if member.value.lower() == value.lower():
                return member
        raise ValueError(
            f"{value} is not a valid {cls.__name__}, must be one of {', '.join([m.value for m in cls])}"
        )
